#include <avr/io.h>
#include "fat1216.h"

uint8_t fat_buf[MMC_SECTOR_SIZE];
uint32_t filesize;

static uint32_t  RootDirRegionStartSec;
static uint32_t  DataRegionStartSec;
static uint16_t  RootDirRegionSize;
static uint8_t   SectorsPerCluster;
static uint32_t  FATRegionStartSec;
static uint8_t   part_type;

static uint16_t currentfatsector;


uint8_t fat1216_init(void)
{
	mbr_t *mbr = (mbr_t*) fat_buf;
	vbr_t *vbr = (vbr_t*) fat_buf;
	uint8_t init, i;
	uint32_t  fss = 0;

//	for (i=0; (i<10) && (init != 0); i++)
		init = mmc_init();
	
	if (init != 0)
		return 1;
		
	//Load MBR
	mmc_read(fss);

	if (mbr->sector.magic != 0xAA55)
		return 2;
		
	//Try sector 0 as MBR
	for (i = 0; i < 4; i++) {
		#ifdef USE_FAT12
		if (((mbr->sector.partition[i].typeId == PARTTYPE_FAT12) || (mbr->sector.partition[i].typeId == PARTTYPE_FAT16L)) && ((mbr->sector.partition[i].state & 0x7f)  == 0))
		#else
		if ((mbr->sector.partition[i].typeId == PARTTYPE_FAT16L) && ((mbr->sector.partition[i].state & 0x7f)  == 0))
		#endif
		{
			fss = mbr->sector.partition[i].sectorOffset;
			//Load VBR
			mmc_read(fss);
			break;
		}
	}
	
	//Check VBR
	if ((vbr->bsFileSysType[0] != 'F') || (vbr->bsFileSysType[3] != '1'))
		return 2;
	
	#ifdef USE_FAT12
	if (vbr->bsFileSysType[4] == '2')
		part_type = PARTTYPE_FAT12;
	else
	#endif
		if (vbr->bsFileSysType[4] == '6')
			part_type = PARTTYPE_FAT16L;
		else
			return 2;
	
	
	SectorsPerCluster  		= vbr->bsSecPerClus;		// 8
	
	// Calculation Algorithms
	fss			+= vbr->bsRsvdSecCnt;						// 6
	RootDirRegionStartSec	= fss + (vbr->bsNumFATs * vbr->bsNrSeProFAT16);		// 496	
	RootDirRegionSize	= vbr->bsRootEntCnt / (MMC_SECTOR_SIZE/sizeof(direntry_t));				// 32
	DataRegionStartSec 	= RootDirRegionStartSec + RootDirRegionSize;	// 528
	FATRegionStartSec	= fss;

	return 0;
}


uint16_t fat1216_readRootDirEntry(uint16_t entry_num)
{
	uint16_t   dirsector;
	direntry_t *dir; //Zeiger auf einen Verzeichniseintrag
	
	dirsector = entry_num / (MMC_SECTOR_SIZE/sizeof(direntry_t));
        if (dirsector >= RootDirRegionSize)
		return 0xFFFF; //End of root dir region reached!
	entry_num %= MMC_SECTOR_SIZE / sizeof(direntry_t);
	mmc_read(RootDirRegionStartSec + dirsector);

	dir = (direntry_t *) fat_buf + entry_num;

	if ((dir->name[0] == 0) || (dir->name[0] == 0xE5) || (dir->fstclust == 0))
		return 0xFFFF;
		
	filesize = dir->filesize;
	
	return dir->fstclust;
}

static void load_fat_sector(uint16_t sec)
{
	if (currentfatsector != sec)
	{
		mmc_read(FATRegionStartSec + sec);
		currentfatsector = sec;
	}
}


void fat1216_readfilesector(uint16_t startcluster, uint16_t filesector)
{
	uint16_t clusteroffset;
	uint8_t  temp, secoffset;
	uint32_t templong;
	
	fatsector_t *fatsector = (fatsector_t*) fat_buf;
	
	clusteroffset = filesector; // seit uint16_t filesector gehts nun doch ;)
	temp = SectorsPerCluster >> 1;
	while(temp)
	{
		clusteroffset >>= 1;
		temp >>= 1;
	}

	secoffset = (uint8_t)filesector & (uint8_t)(SectorsPerCluster-1); // SectorsPerCluster is always power of 2 !
	
	currentfatsector = 0xFFFF;
	
	while (clusteroffset)
	{
		#ifdef USE_FAT12
		if (part_type == PARTTYPE_FAT16L)
		#endif
		{//FAT16
			load_fat_sector(startcluster / 256 );
			startcluster = fatsector->fat16_entries[startcluster % 256];
		}
		
		#ifdef USE_FAT12
		if (part_type == PARTTYPE_FAT12)
		{//FAT12
			uint32_t fat12_bytenum;
			uint8_t clust1, clust2;
	
			//fat12_bytenum = (startcluster * 3 / 2);
			fat12_bytenum = (startcluster + (startcluster>>1));
		
			load_fat_sector(fat12_bytenum / MMC_SECTOR_SIZE);
			clust1 = fatsector->fat_bytes[fat12_bytenum % MMC_SECTOR_SIZE];
			fat12_bytenum++;
			load_fat_sector(fat12_bytenum / MMC_SECTOR_SIZE);
			clust2 = fatsector->fat_bytes[fat12_bytenum % MMC_SECTOR_SIZE];
		
			if (startcluster & (uint8_t)0x01)
				startcluster = ((clust1 & 0xF0) >> 4) | (clust2 << 4);
			else
				startcluster = clust1 | ((clust2 & 0x0F) << 8);
		}
		#endif
		
		clusteroffset--;
	}
	
	if (startcluster < 2)
		return; //Free/reserved Sector
	
	if ((startcluster & 0xFFF0) == 0xFFF0) //FAT16 hat nur 16Bit Clusternummern
		return; //Reserved / bad / last cluster

	templong = startcluster - 2;
	temp = SectorsPerCluster >> 1;
	while(temp)
	{
		templong <<= 1;
		temp >>= 1;
	}
		
	mmc_read(templong + DataRegionStartSec + secoffset);
}
